@@ -1,7 +1,6 @@
 /*
-
 Written by Jeffrey Katcher under contract to Network Appliance.
-Copyright (C) 1997
+Copyright (C) 1997-2001
 Network Appliance, Inc.
 
 This code has been successfully compiled and run by Network
@@ -9,14 +8,16 @@ Appliance on various platforms, including Solaris 2 on an Ultra-170,
 and Windows NT on a Compaq ProLiant. However, this PostMark source
 code is distributed under the Artistic License appended to the end
 of this file. As such, no support is provided. However, please report
-any errors to the author, Jeffrey Katcher <katcher@netcom.com>, or to
+any errors to the author, Jeffrey Katcher <katcher@netapp.com>, or to
 Andy Watson <watson@netapp.com>.
 
 Versions:
 1.00 - Original release - 8/17/97
+
 1.01 - Fixed endless loop on EOF, 
        Divide by zero when file_size_high=file_size_low - 10/29/97
        (Thanks to Chuck Murnane)
+
 1.1 - Added new commands to distribute work across multiple directories
       and/or file systems and multiple work subdirectories.
 
@@ -29,6 +30,24 @@ Versions:
 1.11 - Fixed unfortunate problem where read_file opens in append mode thus
        avoiding actual reads.  (Thanks to Kent Peacock)
 
+1.12 - Changed bytes read and written to float.  Hopefully this will avoid
+       overflow when very large file sizes are used.
+
+1.13 - Added terse report option allowing results to be easily included in
+       other things.  (Thanks to Walter Wong) 
+       Also tweaked help code to allow partial matches
+
+1.14 - Automatically stop run if work files are depleted
+
+1.5 - Many people (most recently Michael Flaster) have emphasized that the
+      pseudo-random number generator was more pseudo than random.  After
+      a review of the literature and extensive benchmarking, I've replaced
+      the previous PRNG with the Mersenne Twister.  While an excellent PRNG,
+      it retains much of the performance of the previous implementation. 
+      URL: http://www.math.keio.ac.jp/~matumoto/emt.html
+      Also changed MB definition to 1024KB, tweaked show command
+
+1.51 - Added command to load config file from CLI
 */
 
 #include <stdio.h>
@@ -37,6 +56,8 @@ Versions:
 #include <time.h>
 #include <fcntl.h>
 
+#define PM_VERSION "v1.51 : 8/14/01"
+
 #ifdef _WIN32
 #include <io.h>
 #include <direct.h>
@@ -56,12 +77,10 @@ extern char *getwd();
 #define MAX_FILENAME 80
 
 #define KILOBYTE 1024
-#define MEGABYTE (1000*KILOBYTE)
+#define MEGABYTE (KILOBYTE*KILOBYTE)
 
 #define PROMPT "pm>"
 
-#define RND(x) ((x>0)?(refrng_rand() % (x)):0)
-
 typedef struct { /* ADT for table of CLI commands */
    char *name;    /* name of command */
    int (*func)(); /* pointer to callback function */
@@ -79,8 +98,10 @@ extern int cli_set_write();
 extern int cli_set_buffering();
 extern int cli_set_bias_read();
 extern int cli_set_bias_create();
+extern int cli_set_report();
 
 extern int cli_run();
+extern int cli_load();
 extern int cli_show();
 extern int cli_help();
 extern int cli_quit();
@@ -99,20 +120,26 @@ cmd command_list[]={ /* table of CLI commands */
       "Sets the chance of choosing read over append"},
    {"set bias create",cli_set_bias_create,
       "Sets the chance of choosing create over delete"},
+   {"set report",cli_set_report,"Choose verbose or terse report format"},
    {"run",cli_run,"Runs one iteration of benchmark"},
+   {"load",cli_load,"Read configuration file"},
    {"show",cli_show,"Displays current configuration"},
    {"help",cli_help,"Prints out available commands"},
    {"quit",cli_quit,"Exit program"},
    NULL
 };
 
+extern void verbose_report();
+extern void terse_report();
+void (*reports[])()={verbose_report,terse_report};
+
 /* Counters */
 int files_created;  /* number of files created */
 int files_deleted;  /* number of files deleted */
 int files_read;     /* number of files read */
 int files_appended; /* number of files appended */
-int bytes_written;  /* number of bytes written to files */
-int bytes_read;     /* number of bytes read from files */
+float bytes_written; /* number of bytes written to files */
+float bytes_read;    /* number of bytes read from files */
 
 /* Configurable Parameters */
 int file_size_low=500;
@@ -126,6 +153,7 @@ int write_block_size=512;
 int bias_read=5;                /* chance of picking read over append */
 int bias_create=5;              /* chance of picking create over delete */
 int buffered_io=1;              /* use C library buffered I/O */
+int report=0;                   /* 0=verbose, 1=terse report format */
 
 /* Working Storage */
 char *file_source; /* pointer to buffer of random text */
@@ -150,22 +178,9 @@ char **location_index;     /* weighted index of file systems */
 
 char *read_buffer; /* temporary space for reading file data into */
 
-static unsigned long int refrng_next=1; /* state variable for random numbers */
-
-/* portable standard random number generator -
-   returns integers from 0 to 2^32 */
-int refrng_rand()
-{
-   refrng_next=refrng_next*1103515245+12345;
-   return((unsigned int)(refrng_next/65536) % 32768);
-}
-
-/* set seed for portable standard random number generator */
-void refrng_srand(seed)
-unsigned int seed;
-{
-   refrng_next=seed;
-}
+#define RND(x) ((x>0)?(genrand() % (x)):0)
+extern unsigned long genrand();
+extern void sgenrand();
 
 /* converts integer values to byte/kilobyte/megabyte strings */
 char *scale(i)
@@ -184,6 +199,23 @@ int i;
    return(buffer);
 }
 
+/* converts float values to byte/kilobyte/megabyte strings */
+char *scalef(i)
+float i;
+{
+   static char buffer[MAX_LINE]; /* storage for current conversion */
+
+   if (i/(float)MEGABYTE>1)
+      sprintf(buffer,"%.2f megabytes",i/(float)MEGABYTE);
+   else
+      if (i/(float)KILOBYTE)
+         sprintf(buffer,"%.2f kilobytes",i/(float)KILOBYTE);
+      else
+         sprintf(buffer,"%f bytes",i);
+
+   return(buffer);
+}
+
 /* UI callback for 'set size' command - sets range of file sizes */
 int cli_set_size(param)
 char *param; /* remainder of command line */
@@ -206,46 +238,40 @@ char *param; /* remainder of command line */
    return(1);
 }
 
+int cli_generic_int(var,param,error)
+int *var; /* pointer to variable to set */
+char *param; /* remainder of command line */
+char *error; /* error message */
+{
+   int value;
+
+   if (param && (value=atoi(param))>0)
+      *var=value;
+   else
+      fprintf(stderr,"Error: %s\n",error);
+
+   return(1);
+}
+
 /* UI callback for 'set number' command - sets number of files to create */
 int cli_set_number(param)
 char *param; /* remainder of command line */
 {
-   int size;
-
-   if (param && (size=atoi(param))>0)
-      simultaneous=size;
-   else
-      fprintf(stderr,"Error: no file number specified\n");
-
-   return(1);
+   return(cli_generic_int(&simultaneous,param,"invalid number of files"));
 }
 
 /* UI callback for 'set seed' command - initial value for random number gen */
 int cli_set_seed(param)
 char *param; /* remainder of command line */
 {
-   int size;
-
-   if (param && (size=atoi(param))>0)
-      seed=size;
-   else
-      fprintf(stderr,"Error: no random number seed specified\n");
-
-   return(1);
+   return(cli_generic_int(&seed,param,"invalid seed for random numbers"));
 }
 
 /* UI callback for 'set transactions' - configure number of transactions */
 int cli_set_transactions(param)
 char *param; /* remainder of command line */
 {
-   int size;
-
-   if (param && (size=atoi(param))>0)
-      transactions=size;
-   else
-      fprintf(stderr,"Error: no transactions specified\n");
-
-   return(1);
+   return(cli_generic_int(&transactions,param,"no transactions specified"));
 }
 
 int parse_weight(params)
@@ -380,42 +406,22 @@ char *param; /* remainder of command line */
 int cli_set_subdirs(param)
 char *param; /* remainder of command line */
 {
-   int subdirs;
-
-   if (param && (subdirs=atoi(param))>=0)
-      subdirectories=subdirs;
-   else
-      fprintf(stderr,"Error: invalid number of subdirectories specified\n");
-
-   return(1);
+   return(cli_generic_int(&subdirectories,param,
+      "invalid number of subdirectories"));
 }
 
 /* UI callback for 'set read' - configure read block size (integer) */
 int cli_set_read(param)
 char *param; /* remainder of command line */
 {
-   int size;
-
-   if (param && (size=atoi(param))>0)
-      read_block_size=size;
-   else
-      fprintf(stderr,"Error: no block size specified\n");
-
-   return(1);
+   return(cli_generic_int(&read_block_size,param,"invalid read block size"));
 }
 
 /* UI callback for 'set write' - configure write block size (integer) */
 int cli_set_write(param)
 char *param; /* remainder of command line */
 {
-   int size;
-
-   if (param && (size=atoi(param))>0)
-      write_block_size=size;
-   else
-      fprintf(stderr,"Error: no block size specified\n");
-
-   return(1);
+   return(cli_generic_int(&write_block_size,param,"invalid write block size"));
 }
 
 /* UI callback for 'set buffering' - sets buffering mode on or off
@@ -437,7 +443,7 @@ char *param; /* remainder of command line */
 {
    int value;
 
-   if (param && (value=atoi(param))>=-1  && value<=10)
+   if (param && (value=atoi(param))>=-1 && value<=10)
       bias_read=value;
    else
       fprintf(stderr,
@@ -452,7 +458,7 @@ char *param; /* remainder of command line */
 {
    int value;
 
-   if (param && (value=atoi(param))>=-1  && value<=10)
+   if (param && (value=atoi(param))>=-1 && value<=10)
       bias_create=value;
    else
       fprintf(stderr,
@@ -461,6 +467,29 @@ char *param; /* remainder of command line */
    return(1);
 }
 
+/* UI callback for 'set report' - chooses verbose or terse report formats */
+int cli_set_report(param)
+char *param; /* remainder of command line */
+{
+   int match=0;
+
+   if (param)
+      { 
+      if (!strcmp(param,"verbose"))
+         report=0;
+      else
+         if (!strcmp(param,"terse"))
+            report=1;
+         else
+            match=-1;
+      }
+
+   if (!param || match==-1)
+      fprintf(stderr,"Error: either 'verbose' or 'terse' required\n");
+
+   return(1);
+}
+
 /* populate file source buffer with 'size' bytes of readable randomness */
 char *initialize_file_source(size)
 int size; /* number of bytes of junk to create */
@@ -487,7 +516,7 @@ time_t t0;
 }
 
 /* prints out results from running transactions */
-void print_report(fp,end_time,start_time,t_end_time,t_start_time,deleted)
+void verbose_report(fp,end_time,start_time,t_end_time,t_start_time,deleted)
 FILE *fp;
 time_t end_time,start_time,t_end_time,t_start_time; /* timers from run */
 int deleted; /* files deleted back-to-back */
@@ -525,10 +554,36 @@ int deleted; /* files deleted back-to-back */
       files_deleted-deleted,(files_deleted-deleted)/t_elapsed);
 
    fprintf(fp,"\nData:\n");
-   fprintf(fp,"\t%s read ",scale(bytes_read));
-   fprintf(fp,"(%s per second)\n",scale(bytes_read/elapsed));
-   fprintf(fp,"\t%s written ",scale(bytes_written));
-   fprintf(fp,"(%s per second)\n",scale(bytes_written/elapsed));
+   fprintf(fp,"\t%s read ",scalef(bytes_read));
+   fprintf(fp,"(%s per second)\n",scalef(bytes_read/(float)elapsed));
+   fprintf(fp,"\t%s written ",scalef(bytes_written));
+   fprintf(fp,"(%s per second)\n",scalef(bytes_written/(float)elapsed));
+}
+
+void terse_report(fp,end_time,start_time,t_end_time,t_start_time,deleted)
+FILE *fp;
+time_t end_time,start_time,t_end_time,t_start_time; /* timers from run */
+int deleted; /* files deleted back-to-back */
+{
+   time_t elapsed,t_elapsed;
+   int interval;
+
+   elapsed=diff_time(end_time,start_time);
+   t_elapsed=diff_time(t_end_time,t_start_time);
+   interval=diff_time(t_start_time,start_time);
+
+   fprintf(fp,"%d %d %.2f ", elapsed, t_elapsed, 
+      (float)transactions/t_elapsed);
+   fprintf(fp, "%.2f %.2f %.2f ", (float)files_created/elapsed, 
+      (float)simultaneous/interval,
+      (float)(files_created-simultaneous)/t_elapsed);
+   fprintf(fp, "%.2f %.2f ", (float)files_read/t_elapsed,
+      (float)files_appended/t_elapsed);
+   fprintf(fp, "%.2f %.2f %.2f ", (float)files_created/elapsed,
+      (float)deleted/interval,
+      (float)(files_deleted-deleted)/t_elapsed);
+   fprintf(fp, "%.2f %.2f\n", (float)bytes_read/elapsed,
+      (float)bytes_written/elapsed);
 }
 
 /* returns file_table entry of unallocated file
@@ -551,7 +606,7 @@ int find_free_file()
    return(-1); /* return -1 only if no free files found */
 }
 
-/* write 'size' bytes to file 'fd' using unbuffered I/O */
+/* write 'size' bytes to file 'fd' using unbuffered I/O and close file */
 void write_blocks(fd,size)
 int fd;
 int size;   /* bytes to write to file */
@@ -567,9 +622,11 @@ int size;   /* bytes to write to file */
    write(fd,file_source+offset,i); /* write remainder */
 
    bytes_written+=size; /* update counter */
+
+   close(fd);
 }
 
-/* write 'size' bytes to file 'fp' using buffered I/O */
+/* write 'size' bytes to file 'fp' using buffered I/O and close file */
 void fwrite_blocks(fp,size)
 FILE *fp;
 int size;   /* bytes to write to file */
@@ -583,8 +640,10 @@ int size;   /* bytes to write to file */
       fwrite(file_source+offset,write_block_size,1,fp);
 
    fwrite(file_source+offset,i,1,fp); /* write remainder */
-
+   
    bytes_written+=size; /* update counter */
+
+   fclose(fp);
 }
 
 void create_file_name(dest)
@@ -633,15 +692,9 @@ int buffered; /* 1=buffered I/O (default), 0=unbuffered I/O */
       if (fp || fd!=-1)
          {
          if (buffered)
-            {
             fwrite_blocks(fp,file_table[free_file].size);
-            fclose(fp);
-            }
          else
-            {
             write_blocks(fd,file_table[free_file].size);
-            close(fd);
-            }
          }
       else
          fprintf(stderr,"Error: cannot open '%s' for writing\n",
@@ -730,15 +783,9 @@ int buffered; /* 1=buffered I/O (default), 0=unbuffered I/O */
          block=RND(file_size_high-file_table[number].size)+1;
 
          if (buffered)
-            {
             fwrite_blocks(fp,block);
-            fclose(fp);
-            }
          else
-            {
             write_blocks(fd,block);
-            close(fd);
-            }
 
          file_table[number].size+=block;
          files_appended++;
@@ -774,7 +821,7 @@ void reset_counters()
 /* perform the configured number of file transactions
    - a transaction consisted of either a read or append and either a
      create or delete all chosen at random */
-void run_transactions(buffered)
+int run_transactions(buffered)
 int buffered; /* 1=buffered I/O (default), 0=unbuffered I/O */
 {
    int percent; /* one tenth of the specified transactions */
@@ -783,6 +830,14 @@ int buffered; /* 1=buffered I/O (default), 0=unbuffered I/O */
    percent=transactions/10;
    for (i=0; i<transactions; i++)
       {
+      if (files_created==files_deleted)
+         {
+         printf("out of files!\n");
+         printf("For this workload, either increase the number of files or\n");
+         printf("decrease the number of transactions.\n");
+         break;
+         }
+
       if (bias_read!=-1) /* if read/append not locked out... */
          {
          if (RND(10)<bias_read) /* read file */
@@ -805,6 +860,8 @@ int buffered; /* 1=buffered I/O (default), 0=unbuffered I/O */
          fflush(stdout);
          }
       }
+
+   return(transactions-i);
 }
 
 char **build_location_index(list,weight)
@@ -890,11 +947,12 @@ char *param; /* unused */
    time_t start_time,t_start_time,t_end_time,end_time; /* elapsed timers */
    int delete_base; /* snapshot of deleted files counter */
    FILE *fp=NULL; /* file descriptor for directing output */
+   int incomplete;
    int i; /* generic iterator */
 
    reset_counters(); /* reset counters before each run */
 
-   refrng_srand(seed); /* initialize random number generator */
+   sgenrand(seed); /* initialize random number generator */
 
    /* allocate file space and fill with junk */
    file_source=initialize_file_source(file_size_high<<1);
@@ -933,9 +991,10 @@ char *param; /* unused */
    printf("Performing transactions");
    fflush(stdout);
    time(&t_start_time);
-   run_transactions(buffered_io);
+   incomplete=run_transactions(buffered_io);
    time(&t_end_time);
-   printf("Done\n");
+   if (!incomplete)
+      printf("Done\n");
 
    /* delete remaining files */
    printf("Deleting files...");
@@ -970,8 +1029,9 @@ char *param; /* unused */
    if (!fp)
       fp=stdout;
 
-   print_report(fp,end_time,start_time,t_end_time,t_start_time,
-      files_deleted-delete_base);
+   if (!incomplete)
+      reports[report](fp,end_time,start_time,t_end_time,t_start_time,
+         files_deleted-delete_base);
 
    if (param && fp!=stdout)
       fclose(fp);
@@ -984,6 +1044,20 @@ char *param; /* unused */
    return(1); /* return 1 unless exit requested, then return 0 */
 }
 
+/* CLI callback for 'load' - read configuration file */
+int cli_load(param)
+char *param;
+{
+   char buffer[MAX_LINE+1]; /* storage for input command line */
+
+   if (param)
+      read_config_file(param,buffer,0);
+   else
+      fprintf(stderr,"Error: no configuration file specified\n");
+
+   return(1); /* return 1 unless exit requested, then return 0 */
+}
+
 /* CLI callback for 'show' - print values of configuration variables */
 int cli_show(param) 
 char *param; /* optional: name of output file */
@@ -1000,24 +1074,35 @@ char *param; /* optional: name of output file */
       fp=stdout;
 
    fprintf(fp,"Current configuration is:\n");
-   fprintf(fp,"Transactions: %d\n",transactions);
-   fprintf(fp,"Files range between %s ",scale(file_size_low));
-   fprintf(fp,"and %s in size\n",scale(file_size_high));
-   fprintf(fp,"Random number generator seed is %d\n",seed);
    fprintf(fp,"The base number of files is %d\n",simultaneous);
+   fprintf(fp,"Transactions: %d\n",transactions);
+
+   if (file_size_low!=file_size_high)
+      {
+      fprintf(fp,"Files range between %s ",scale(file_size_low));
+      fprintf(fp,"and %s in size\n",scale(file_size_high));
+      }
+   else
+      fprintf(fp,"Files are %s in size\n",scale(file_size_low));
+
    fprintf(fp,"Working director%s: %s\n",(file_system_count>1)?"ies":"y",
       (file_system_count==0)?GETWD(current_dir):"");
 
    for (traverse=file_systems; traverse; traverse=traverse->next)
       printf("\t%s (weight=%d)\n",traverse->system.name,traverse->system.size);
 
-   fprintf(fp,"%d subdirector%s will be used\n",subdirectories,
-      (subdirectories==1)?"y":"ies");
+   if (subdirectories>0)
+      fprintf(fp,"%d subdirector%s will be used\n",subdirectories,
+         (subdirectories==1)?"y":"ies");
+
    fprintf(fp,"Block sizes are: read=%s, ",scale(read_block_size));
    fprintf(fp,"write=%s\n",scale(write_block_size));
    fprintf(fp,"Biases are: read/append=%d, create/delete=%d\n",bias_read,
       bias_create);
    fprintf(fp,"%ssing Unix buffered file I/O\n",buffered_io?"U":"Not u");
+   fprintf(fp,"Random number generator seed is %d\n",seed);
+
+   fprintf(fp,"Report format is %s.\n",report?"terse":"verbose");
 
    if (param && fp!=stdout)
       fclose(fp);
@@ -1036,17 +1121,19 @@ char *param; /* unused */
 int cli_help(param) 
 char *param; /* optional: specific command to get help for */
 {
+   int n=0; /* number of matching items */
    int i; /* traversal variable for command table */
+   int len;
 
-   if (param) /* if a command is specified... */
+   if (param && (len=strlen(param))>0) /* if a command is specified... */
       for (i=0; command_list[i].name; i++) /* walk command table */
-         if (!strcmp(command_list[i].name,param))
+         if (!strncmp(command_list[i].name,param,len))
             {
             printf("%s - %s\n",command_list[i].name,command_list[i].help);
-            break;
+            n++;
             }
 
-   if (!param || !command_list[i].name)
+   if (!param || !n)
       for (i=0; command_list[i].name; i++) /* traverse command table */
          printf("%s - %s\n",command_list[i].name,command_list[i].help);
 
@@ -1104,9 +1191,10 @@ char *buffer; /* line of user input */
 
 /* read config file if present and process it line by line
    - if 'quit' is in file then function returns 0 */
-int read_config_file(filename,buffer)
+int read_config_file(filename,buffer,ignore)
 char *filename; /* file name of config file */
 char *buffer;   /* temp storage for each line read from file */
+int ignore;     /* ignore file not found */
 {
    int result=1; /* default exit value - proceed with UI */
    FILE *fp;
@@ -1122,19 +1210,23 @@ char *buffer;   /* temp storage for each line read from file */
 
       fclose(fp);
       }
+   else
+      if (!ignore)
+         fprintf(stderr,"Error: cannot read configuration file '%s'\n",
+            filename); 
 
    return(result);
 }
 
 /* main function - reads config files then enters get line/parse line loop */
-void main(argc,argv)
+main(argc,argv)
 int argc;
 char *argv[];
 {
    char buffer[MAX_LINE+1]; /* storage for input command line */
 
-   printf("PostMark v1.11 : 4/8/98\n");
-   if (read_config_file((argc==2)?argv[1]:".pmrc",buffer))
+   printf("PostMark %s\n",PM_VERSION);
+   if (read_config_file((argc==2)?argv[1]:".pmrc",buffer,1))
       while (cli_read_line(buffer,MAX_LINE) && cli_parse_line(buffer))
          ;
 }
@@ -1270,3 +1362,130 @@ WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
                                 The End
 
 */
+
+
+/* A C-program for MT19937: Integer version (1999/10/28)          */
+/*  genrand() generates one pseudorandom unsigned integer (32bit) */
+/* which is uniformly distributed among 0 to 2^32-1  for each     */
+/* call. sgenrand(seed) sets initial values to the working area   */
+/* of 624 words. Before genrand(), sgenrand(seed) must be         */
+/* called once. (seed is any 32-bit integer.)                     */
+/*   Coded by Takuji Nishimura, considering the suggestions by    */
+/* Topher Cooper and Marc Rieffel in July-Aug. 1997.              */
+
+/* This library is free software; you can redistribute it and/or   */
+/* modify it under the terms of the GNU Library General Public     */
+/* License as published by the Free Software Foundation; either    */
+/* version 2 of the License, or (at your option) any later         */
+/* version.                                                        */
+/* This library is distributed in the hope that it will be useful, */
+/* but WITHOUT ANY WARRANTY; without even the implied warranty of  */
+/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            */
+/* See the GNU Library General Public License for more details.    */
+/* You should have received a copy of the GNU Library General      */
+/* Public License along with this library; if not, write to the    */
+/* Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA   */ 
+/* 02111-1307  USA                                                 */
+
+/* Copyright (C) 1997, 1999 Makoto Matsumoto and Takuji Nishimura. */
+/* Any feedback is very welcome. For any question, comments,       */
+/* see http://www.math.keio.ac.jp/matumoto/emt.html or email       */
+/* matumoto@math.keio.ac.jp                                        */
+
+/* REFERENCE                                                       */
+/* M. Matsumoto and T. Nishimura,                                  */
+/* "Mersenne Twister: A 623-Dimensionally Equidistributed Uniform  */
+/* Pseudo-Random Number Generator",                                */
+/* ACM Transactions on Modeling and Computer Simulation,           */
+/* Vol. 8, No. 1, January 1998, pp 3--30.                          */
+
+/* Period parameters */  
+#define N 624
+#define M 397
+#define MATRIX_A 0x9908b0df   /* constant vector a */
+#define UPPER_MASK 0x80000000 /* most significant w-r bits */
+#define LOWER_MASK 0x7fffffff /* least significant r bits */
+
+/* Tempering parameters */   
+#define TEMPERING_MASK_B 0x9d2c5680
+#define TEMPERING_MASK_C 0xefc60000
+#define TEMPERING_SHIFT_U(y)  (y >> 11)
+#define TEMPERING_SHIFT_S(y)  (y << 7)
+#define TEMPERING_SHIFT_T(y)  (y << 15)
+#define TEMPERING_SHIFT_L(y)  (y >> 18)
+
+static unsigned long mt[N]; /* the array for the state vector  */
+static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */
+
+/* Initializing the array with a seed */
+void
+sgenrand(seed)
+    unsigned long seed;	
+{
+    int i;
+
+    for (i=0;i<N;i++) {
+         mt[i] = seed & 0xffff0000;
+         seed = 69069 * seed + 1;
+         mt[i] |= (seed & 0xffff0000) >> 16;
+         seed = 69069 * seed + 1;
+    }
+    mti = N;
+}
+
+/* Initialization by "sgenrand()" is an example. Theoretically,      */
+/* there are 2^19937-1 possible states as an intial state.           */
+/* This function allows to choose any of 2^19937-1 ones.             */
+/* Essential bits in "seed_array[]" is following 19937 bits:         */
+/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[N-1]. */
+/* (seed_array[0]&LOWER_MASK) is discarded.                          */ 
+/* Theoretically,                                                    */
+/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[N-1]  */
+/* can take any values except all zeros.                             */
+void
+lsgenrand(seed_array)
+    unsigned long seed_array[]; 
+    /* the length of seed_array[] must be at least N */
+{
+    int i;
+
+    for (i=0;i<N;i++) 
+      mt[i] = seed_array[i];
+    mti=N;
+}
+
+unsigned long 
+genrand()
+{
+    unsigned long y;
+    static unsigned long mag01[2]={0x0, MATRIX_A};
+    /* mag01[x] = x * MATRIX_A  for x=0,1 */
+
+    if (mti >= N) { /* generate N words at one time */
+        int kk;
+
+        if (mti == N+1)   /* if sgenrand() has not been called, */
+            sgenrand(4357); /* a default initial seed is used   */
+
+        for (kk=0;kk<N-M;kk++) {
+            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
+            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
+        }
+        for (;kk<N-1;kk++) {
+            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
+            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
+        }
+        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
+        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];
+
+        mti = 0;
+    }
+  
+    y = mt[mti++];
+    y ^= TEMPERING_SHIFT_U(y);
+    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
+    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
+    y ^= TEMPERING_SHIFT_L(y);
+
+    return y; 
+}
